/*
*  C Implementation: BoardController
*
* Description: Implementation of the board controller (LED's, Switches ...)
*
*
* Author: Michel Tokic <michel@tokic.com>, (C) 2009
*
* Copyright: See COPYING file that comes with this distribution
* 
*/

#include "BoardController.h"

/** initialize ports for LEDs, switches and the serial interface.*/
void BoardController::initPorts() {

	// initialize switch-pins as inputs (PA0-7)
	DDRA = 0x00;
	// enable internal pull-ups for switch-pins (PA0-7)
	PORTA |= (1<<PA0) | (1<<PA1) | (1<<PA2) | (1<<PA3) | (1<<PA4) | (1<<PA5) | (1<<PA6) | (1<<PA7);


	// initialize LED pins as outputs (PB0-3)
	DDRB = (1<<PB0) | (1<<PB1) | (1<<PB2) | (1<<PB3); // LED1..4

	// turn on power led
	powerOnLED(LED_POWER, true);
	powerOnLED(LED_WALK, true);
	powerOnLED(LED_LEARN, true);
	powerOnLED(LED_EXPLORATION, true);

	msDelay(800);

	// LED test (turn all LEDs off)
	powerOnLED(LED_POWER, true);
	powerOnLED(LED_WALK, false);
	powerOnLED(LED_LEARN, false);
	powerOnLED(LED_EXPLORATION, false);



	// define serial interface control pins as outputs ... 
	DDRC |= (1<<EN_TXD_BIOLOID) | (1<<EN_RXD_BIOLOID) | (1<<EN_AIR_V24) | (1<<RTS);

	// disable serial port
	UCSRB &= ~((1<<TXEN) | (1<<RXEN));   // disable TX and RX

	// define RTS as input
	DDRC &= ~(1<<CTS);
	// indicate Host that serial line is not ready for reception
	PORTC |= (1<<RTS);

	// USART pins
	DDRD |= (1<<PD1);
	DDRD &= ~(1<<PD0);
}

/** return the switch status !! low-active pins are converted to: On:=1, Off=0) */
unsigned char BoardController::getSwitchState (unsigned char _switch) {

	bool status;

	if (_switch == SWITCH_WALK) {
		status = (PINA & (1<<PA0)); // Bioloid
		
	} else if (_switch == SWITCH_RANDOM_STATE) {
		status = (PINA & (1<<PA1)); // Bioloid
		
	} else if (_switch == SWITCH_INV_FB_0) {
		status = (PINA & (1<<PA2)); // Bioloid
	
	} else if (_switch == SWITCH_ENABLE_EXPLORATION) {
		status = (PINA & (1<<PA3)); // Bioloid
		
	} else if (_switch == SWITCH_FULL_EXP_AT_START) {
		status = (PINA & (1<<PA4)); // PCBv2 and Bioloid
		
	} else if (_switch == SWITCH_V24) {
		status = (PINA & (1<<PA5)); // PCBv2 and Bioloid 
						
	} else if (_switch == SWITCH_VDBE_EGREEDY) {
		status = (PINA & (1<<PA6)); // PCBv2 and Bioloid

	} else if (_switch == SWITCH_GAMMA) {
		status = (PINA & (1<<PA7)); // PCBv2 and Bioloid

	} else if (_switch == SWITCH_RAM) {
		status = SWITCH_RAM_VALUE; // PCBv2

	// blink if _switch is unknown
	} else {
		
		register char i;
		
		for (i=0; i<100; i++) {
		
			powerOnLED(LED_POWER, 1);
			powerOnLED(LED_WALK, 0);
			powerOnLED(LED_LEARN, 0);
			powerOnLED(LED_EXPLORATION, 1);
			
			msDelay(250);
			
			powerOnLED(LED_POWER, 0);
			powerOnLED(LED_WALK, 1);
			powerOnLED(LED_LEARN, 1);
			powerOnLED(LED_EXPLORATION, 0);
			
			msDelay(250);
		}
		
		return 0;
	}
	
	return !status;
}

/** enable serial interrupt */
void BoardController::enableV24Interrupt()  { 

	disableV24Interrupt();
 	if (getSwitchState(SWITCH_V24)) {
 		//UCSRB |=  (1 << RXCIE); 
 		//sei();
 	}
	//PORTC &= ~(1<<PC3);
	//UCSRB |= (1<<RXEN);
}

/** disable serial interrupt  */
void BoardController::disableV24Interrupt() { 
	if (getSwitchState(SWITCH_V24)) {
		//UCSRB &= ~(1 << RXCIE); 
		//cli();
	}
	//PORTC  |= (1<<PC3);
	//PORTC |= (1<<PC3);
	//UCSRB &= ~(1<<RXEN);
}


/** delay of approx. n milli-seconds */
void BoardController::msDelay (int n) {

	//double maxDelay = (262.14 / (double)((11059200UL / 1000000)));
	double maxDelay = (262.14 / (double)((16000000UL / 1000000)));

	for (int i=0; i< (n/maxDelay); i++) {
		_delay_ms (maxDelay);
	}
}


/** turns a specific LED on (_status=true) of off (_status=false) */
void BoardController::powerOnLED(unsigned char _ledID, bool _status) {

	if (_status == true) {
		PORTB &= ~(1<<_ledID);
	} else {
		PORTB |= (1<<_ledID);
	}
}


void BoardController::blinkLED (unsigned char led_id, unsigned int count, unsigned int delay) {

	unsigned char i;
// 	DDRB = 0x0F;
// 	PORTB |= (1<<PA4) | (1<<PA5) | (1<<PA6) | (1<<PA7);

	for (i=0; i<count; i++) {

 		if (led_id & (1<<LED_WALK)) powerOnLED ( LED_WALK, true);
 		if (led_id & (1<<LED_LEARN)) powerOnLED ( LED_LEARN, true);
 		if (led_id & (1<<LED_EXPLORATION)) powerOnLED ( LED_EXPLORATION, true);
 		if (led_id & (1<<LED_POWER)) powerOnLED ( LED_POWER, true);

		msDelay (delay);

 		if (led_id & (1<<LED_LEARN)) powerOnLED ( LED_LEARN, false);
 		if (led_id & (1<<LED_WALK)) powerOnLED ( LED_WALK, false);
 		if (led_id & (1<<LED_EXPLORATION)) powerOnLED ( LED_EXPLORATION, false);
 		if (led_id & (1<<LED_POWER)) powerOnLED ( LED_POWER, true);

		msDelay (delay);
	}
}


/** setup the serial interface */
void BoardController::enableSerialBioloidInterface() {

	// indicate Host that serial line is not ready for reception
	PORTC  |= (1<<RTS);

	// disable serial wires
	PORTC  &= ~((1<<EN_TXD_BIOLOID) | (1<<EN_RXD_BIOLOID) | (1<<EN_AIR_V24));

	// flushing the buffer
	unsigned char dummy;
	while (UCSRA & (1<<RXC)) dummy = UDR;

	// ** Serial init **
	// 1) USART
	//UCSRB &= ~((1<<TXEN) | (1<<RXEN));   // disable TX and RX

	// set baudrate 
	UBRRH = (unsigned char) (BIOLOID_UBRR >> 8);  // load the upper bits
	UBRRL = (unsigned char) BIOLOID_UBRR; 

	UCSRA |= (1<<U2X);
	UCSRB |= ((1<<TXEN) | (1<<RXEN));   // enable transmit and receive.enable TX and RX
	//UCSRC = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0); // Frame Format: 8data, 1stop bit, noParity 

	// enable bioloid transmission pin
	PORTC  |= (1<<EN_TXD_BIOLOID);
	PORTC  &= ~((1<<EN_RXD_BIOLOID) | (1<<EN_AIR_V24));

}

void BoardController::enableSerialV24Interface() {

	BoardController::powerOnLED(LED_EXPLORATION, true);

	// disable serial wires
	PORTC  &= ~((1<<EN_TXD_BIOLOID) | (1<<EN_RXD_BIOLOID) | (1<<EN_AIR_V24));

	// flushing the buffer
	unsigned char dummy;
	while (UCSRA & (1<<RXC)) dummy = UDR;

	// ** Serial init **
	// 1) USART
	//UCSRB &= ~((1<<TXEN) | (1<<RXEN));   // disable TX and RX

	// set baudrate 
	UBRRH = (unsigned char) (V24_UBRR >> 8);  // load the upper bits
	UBRRL = (unsigned char) V24_UBRR; 

	UCSRA |= (1<<U2X);
	UCSRB = ((1<<TXEN) | (1<<RXEN));   // enable transmit and receive.enable TX and RX

	// disable bioloid bus and enable V24
	PORTC  |= (1<<EN_AIR_V24);
	PORTC  &= ~((1<<EN_TXD_BIOLOID) | (1<<EN_RXD_BIOLOID));

	// ... ready to receive
	PORTC  &= ~(1<<RTS);

	BoardController::msDelay(50);
	BoardController::powerOnLED(LED_EXPLORATION, false);
}

void BoardController::disableSerialInterfaces() {

	//PORTC  &= ~(1<<CTS);

	// disable serial wires
	PORTC  &= ~((1<<EN_TXD_BIOLOID) | (1<<EN_RXD_BIOLOID) | (1<<EN_AIR_V24));
	UCSRB &= ~((1<<TXEN) | (1<<RXEN));   // disable TX and RX

	if (getSwitchState(SWITCH_V24)) {
		enableSerialV24Interface();
	}
}


