

#include <avr/io.h>
#include <avr/interrupt.h>
#include "SensorOnlyMode.h"

// convert a word into 2 bytes
#define lowbyte(w)  ((byte) w)
#define highbyte(w) ((byte) (w >> 8))


void stm_init(Statemachine *sm, byte id, void (*callback)(Packet *pkt))
{
  sm->state = STATE_IDLE;
  sm->id = id;
  sm->callback = callback;
}

void stm_process(Statemachine *sm,  byte b )
{
  switch(sm->state)
  {
    case STATE_IDLE:
    {
      if( b == 0xFF )
        sm->state = STATE_1ST_FF_RCVD;
      break;
    }
      
    case STATE_1ST_FF_RCVD:
    {
      if( b == 0xFF )
        sm->state = STATE_2ND_FF_RCVD;
      else
        sm->state = STATE_IDLE;
      break;      
    }
      
    case STATE_2ND_FF_RCVD:
    {
      if( b != sm->id && b != BROADCAST_ID){
        sm->state = STATE_IDLE;
        break;
      }
      
      sm->state = STATE_ID_RCVD;
      sm->pkt.id  = b;
      sm->pkt.crc = b;
      break;
    }
      
    case STATE_ID_RCVD:
    {
      if( b > MAX_PACKELEN ){
        sm->state = STATE_IDLE;
        break;
      }
      
      sm->pkt.len = b;
      sm->pkt.crc += b;
      sm->state = STATE_LENGTH_RCVD;
      break;
    }
      
    case STATE_LENGTH_RCVD:
    {
      sm->pkt.paramPtr = 0;
      sm->pkt.cmd = b;
      sm->pkt.crc += b;
      sm->state = STATE_COMMAND_RCVD;
      break;
    }
      
    case STATE_COMMAND_RCVD:
    {
      if( sm->pkt.paramPtr+2 >= sm->pkt.len ){
        sm->state = STATE_PKT_COMPLETE;
      }
      else {
        sm->pkt.param[ sm->pkt.paramPtr++ ] = b;
        sm->pkt.crc += b;
        break;
      }
    }
      
    case STATE_PKT_COMPLETE:
    {
      sm->state = STATE_IDLE;    
      sm->pkt.crc += b;      
      sm->callback( &sm->pkt );
      break;
    }
    default:
      sm->state = STATE_IDLE;
  }
}



volatile Smartbuffer interruptbuff;
static byte param[ MAX_PARAM ];
static signed int wheel1, wheel2;

byte TxPacket(byte bError, byte bParameterLength);
void TxD(byte bTxdData);
void callback( Packet *pkt );
void initAtmega(void);

signed char dist;

// Reward Sensor object
FeedbackSensor *fs = new FeedbackSensor();

//int sensorMain (void) 
int sensorMain (void) 
{

  Statemachine stm;
  stm_init(&stm, SELF_ID, &callback);
  RESET_SENSOR;
  init(interruptbuff);
  initAtmega();
  
  RS485_RXD;
  for(;;){
  	if( !isEmpty(interruptbuff) ){
    	stm_process(&stm, pop(interruptbuff) );
    }  
  }
}



void initAtmega(void)
{
  // init
	DDRB   = 0x00; // setup PortB 
	TCCR0  = 0x07; // inic count0
	TCNT0  = 0x00;
	TCNT1L = 0x00; // inic count1
	TCNT1H = 0x00;
	TCCR1A = 0x00;
	TCCR1B = 0x07;				
  
  
	// UART
	UCSRA = (1<<U2X);
	UBRRL = XTAL / (BAUD*8L) - 1;               // speed
	UBRRH = 0;                                  // load the upper bits
	UCSRB = (1<<TXEN) | (1<<RXEN) | (1<<RXCIE);	// enable transmit and receive.enable TX and RX	
  
  
	// wheel
	// Data A is on PB0
	// Data B is on PD2 (INT0)
	DDRD   = 0x00;
	DDRD  |= (1<<PD1) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7);
	DDRB  &= ~(1<<PB0);
	DDRB  &= ~(1<<PB1);
	MCUCR |= (1<<ISC10) | (1<<ISC00);
	
	// interupt enable for PD2, PD3
	GICR |= (1<<INT0) | (1<<INT1);
	GIFR |= (1<<INTF0) | (1<<INTF1);
	
	sei();  //Enable Interrupt   
	RS485_RXD;
}

/* This functions transmit a packet */
byte TxPacket(byte bError, byte bParameterLength) 
{
  byte bCount,bCheckSum,bPacketLength;
  byte txbuff[ MAX_PACKELEN ];
  
  txbuff[0] = 0xff;
  txbuff[1] = 0xff;
  txbuff[2] = SELF_ID;
  txbuff[3] = bParameterLength+2; // Length(Paramter,Instruction,Checksum)
  txbuff[4] = bError;
  
  for(bCount = 0; bCount < bParameterLength; bCount++)
    txbuff[bCount+5] = param[bCount];
  
  bCheckSum = 0;
  bPacketLength = bParameterLength+4+2;
  
  for(bCount = 2; bCount < bPacketLength-1; bCount++) // except 0xff,checksum
    bCheckSum += txbuff[bCount];
  
  txbuff[bCount] = ~bCheckSum; // Writing Checksum with Bit Inversion
  
  RS485_TXD;
  for(bCount = 0; bCount < bPacketLength; bCount++)
  {
    SET_TXD_FINISH;
    TxD(txbuff[bCount]);
  }
  while(!CHECK_TXD_FINISH); // Wait until TXD Shift register empty
  RS485_RXD;
  
  return(bPacketLength);
}

void TxD(byte bTxdData)
{
  while(!TXD_READY);
  TXD_DATA = bTxdData;  
}


void callback( Packet *pkt )
{  
  // checksum test
  if( pkt->crc != 0xff ){
    TxPacket(E_CHECKSUM,0);
    return;
  }   
  
  // ping
  if( pkt->cmd == INST_PING ){
    TxPacket(NO_ERROR,0);
    return;
  }  
  
  // read
  if( pkt->cmd == INST_READ && pkt->param[0] == P_DIST ){
    
    // check the limits of the size
    if( wheel1 > WORD_MAX_SIZE ) wheel1 = WORD_MAX_SIZE;
    if( wheel1 < WORD_MIN_SIZE ) wheel1 = WORD_MIN_SIZE;
    if( wheel2 > WORD_MAX_SIZE ) wheel2 = WORD_MAX_SIZE;
    if( wheel2 < WORD_MIN_SIZE ) wheel2 = WORD_MIN_SIZE;
    
    // send a signed word for both wheels
    param[0] = lowbyte(wheel1);
    param[1] = highbyte_s(wheel1);
    param[2] = lowbyte(wheel2);
    param[3] = highbyte_s(wheel2);
    TxPacket(NO_ERROR,4);
    return;
  }
  
  // reset
  if( pkt->cmd == INST_RESET ){
    TxPacket(NO_ERROR,0);
    RESET_WHEEL;
    return;
  }                    
}

