#include "global.h"
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/io.h>

#define MAX_PARAM       8
#define MAX_PACKELEN    MAX_PARAM+4
#define BROADCAST_ID    0xFE
#define SELF_ID         200

//--- ATMEGA ---
#define ENABLE_BIT_DEFINITIONS 
#define cbi(REG,BITNUM) REG &= ~(_BV(BITNUM)) // clear bit
#define sbi(REG,BITNUM) REG |= _BV(BITNUM)    // set bit

#define WORD_MAX_SIZE  32767
#define WORD_MIN_SIZE -32767

// convert a word into 2 bytes
#define lowbyte(w)  ((byte) w)
#define highbyte(w) ((byte) (w >> 8))

// convert a word into 2 bytes with sign
#define highbyte_s(w) (w&0x80000000?(byte)(w >> 8)|(0x8000):(byte)(w >> 8))

//--- ERRORCODES ---
#define NO_ERROR        0
#define E_INPUT_VOLTAGE 1
#define E_ANGLE_LIMIT   2
#define E_OVERHEATING   4
#define E_RANGE         8
#define E_CHECKSUM      16
#define E_OVERLOAD      32
#define E_INSTRUCTION   64

//---INSTRUCTIONS ---
#define INST_PING           0x01
#define INST_READ           0x02
#define INST_WRITE          0x03
#define INST_REG_WRITE      0x04
#define INST_ACTION         0x05
#define INST_RESET          0x06
#define INST_DIGITAL_RESET  0x07
#define INST_SYSTEM_READ    0x0C
#define INST_SYSTEM_WRITE   0x0D
#define INST_SYNC_WRITE     0x83
#define INST_SYNC_REG_WRITE 0x84

//--FLASH--
#define P_DIST 1

typedef unsigned char byte;

/***********************************************************************
* PACKET
***********************************************************************/

typedef struct
{
	byte id;
	byte len;
	byte cmd;
	byte paramPtr;
	byte param[ MAX_PARAM ];
	byte crc; 
} BioloidPacket;

/***********************************************************************
* STATEMACHINE
***********************************************************************/

#define STATE_IDLE          0   
#define STATE_1ST_FF_RCVD   1  
#define STATE_2ND_FF_RCVD   2   
#define STATE_ID_RCVD       3  
#define STATE_LENGTH_RCVD   4   
#define STATE_COMMAND_RCVD  5   
#define STATE_PKT_COMPLETE  6   

typedef struct 
{
	byte id;
	byte state;
	BioloidPacket pkt;
	void (*callback)(BioloidPacket *pkt);
} Statemachine;

void stm_init(Statemachine *sm, byte id, void (*callback)(BioloidPacket *pkt))
{
	sm->state = STATE_IDLE;
	sm->id = id;
	sm->callback = callback;
}

void stm_process(Statemachine *sm,  byte b )
{
	switch(sm->state)
	{
		case STATE_IDLE:
		{
			if( b == 0xFF )
				sm->state = STATE_1ST_FF_RCVD;
			break;
		}
      
		case STATE_1ST_FF_RCVD:
		{
			if( b == 0xFF )
				sm->state = STATE_2ND_FF_RCVD;
			else
				sm->state = STATE_IDLE;
			break;      
		}
      
		case STATE_2ND_FF_RCVD:
		{
			if( b != sm->id && b != BROADCAST_ID){
				sm->state = STATE_IDLE;
				break;
			}
      
			sm->state = STATE_ID_RCVD;
			sm->pkt.id  = b;
			sm->pkt.crc = b;
			break;
		}
      
		case STATE_ID_RCVD:
		{
			if( b > MAX_PACKELEN ){
				sm->state = STATE_IDLE;
				break;
			}
      
			sm->pkt.len = b;
			sm->pkt.crc += b;
			sm->state = STATE_LENGTH_RCVD;
			break;
		}
      
		case STATE_LENGTH_RCVD:
		{
			sm->pkt.paramPtr = 0;
			sm->pkt.cmd = b;
			sm->pkt.crc += b;
			sm->state = STATE_COMMAND_RCVD;
			break;
		}
      
		case STATE_COMMAND_RCVD:
		{
			if( sm->pkt.paramPtr+2 >= sm->pkt.len ){
				sm->state = STATE_PKT_COMPLETE;
			}
			else {
				sm->pkt.param[ sm->pkt.paramPtr++ ] = b;
				sm->pkt.crc += b;
				break;
			}
		}
      
		case STATE_PKT_COMPLETE:
		{
			sm->state = STATE_IDLE;    
			sm->pkt.crc += b;      
			sm->callback( &sm->pkt );
			break;
		}
		default:
			sm->state = STATE_IDLE;
	}
}


/***********************************************************************
* SMARTBUFFER
***********************************************************************/

#define init(sb)        (sb.readPtr = sb.writePtr = 0)
#define isEmpty(sb)     (sb.writePtr == sb.readPtr)
#define push(sb, value) (sb.buffer[ sb.writePtr++ ] = value)
#define pop(sb)         (sb.buffer[ sb.readPtr++ ])
#define len(sb) ( sb.readPtr > sb.writePtr ? (sb.writePtr - sb.readPtr + 256) : (sb.writePtr - sb.readPtr) )

typedef struct {
	volatile byte  readPtr;
	volatile byte  writePtr;
	byte     buffer[ 256 ];
} Smartbuffer;



/***********************************************************************
* MAIN
***********************************************************************/

#define RS485_TXD sbi(PORTD, PD5), cbi(PORTD, PD6), cbi(UCSRB,RXCIE)
#define RS485_RXD sbi(PORTD, PD6), cbi(PORTD, PD5), sbi(UCSRB,RXCIE)

#define SET_TXD_FINISH   sbi(UCSRA,TXC)
#define RESET_TXD_FINISH cbi(UCSRA,TXC)
#define CHECK_TXD_FINISH bit_is_set(UCSRA,TXC)

#define TXD_READY			bit_is_set(UCSRA,UDRE)
#define TXD_DATA			UDR
#define RXD_DATA			UDR

#define READDIST_X  TCNT0
#define READDIST_Y  TCNT1
#define RESET_MOUSE TCNT0 = TCNT1 = 0
#define RESET_WHEEL wheel1 = wheel2 = 0
#define RESET_SENSOR RESET_WHEEL

//#define XTAL 11059000
//#define BAUD 9600

#define XTAL 16000000
#define BAUD 1000000

static Smartbuffer interruptbuff;
static byte param[ MAX_PARAM ];
signed int wheel1;
signed int wheel2;

byte TxPacket(byte bError, byte bParameterLength);
void TxD(byte bTxdData);
void callback( BioloidPacket *pkt );
void initAtmega(void);

signed char dist;

// int main (void) 
// {
// 	Statemachine stm;
// 	stm_init(&stm, SELF_ID, &callback);
// 	RESET_SENSOR;
// 	init(interruptbuff);
// 	initAtmega();
//   
// 	RS485_RXD;
// 	for(;;){
// 		if( !isEmpty(interruptbuff) ){
// 			stm_process(&stm, pop(interruptbuff) );
// 		}  
// 	}
// }



void initAtmega(void)
{
  // init
	DDRB   = 0x00; // setup PortB 
	TCCR0  = 0x07; // inic count0
	TCNT0  = 0x00;
	TCNT1L = 0x00; // inic count1
	TCNT1H = 0x00;
	TCCR1A = 0x00;
	TCCR1B = 0x07;				
  
  
	// UART
	UCSRA = (1<<U2X);
	UBRRL = XTAL / (BAUD*8L) - 1;               // speed
	UBRRH = 0;                                  // load the upper bits
	UCSRB = (1<<TXEN) | (1<<RXEN) | (1<<RXCIE);	// enable transmit and receive.enable TX and RX	
  
  
  // wheel
  // Data A is on PB0
  // Data B is on PD2 (INT0)
	DDRD   = 0x00;
	DDRD  |= (1<<PD1) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7);
	DDRB  &= ~(1<<PB0);
	DDRB  &= ~(1<<PB1);
	MCUCR |= (1<<ISC10) | (1<<ISC00);
  
  // interupt enable for PD2, PD3
	GICR |= (1<<INT0) | (1<<INT1);
	GIFR |= (1<<INTF0) | (1<<INTF1);
  
	sei();  //Enable Interrupt   
	RS485_RXD;
}

SIGNAL(SIG_INTERRUPT0)
{ 
	register unsigned char pd = PIND;
	register unsigned char pb = PINB;
  
  // if the signals are different is was
  // a forwad movement - else a backward movement
	if(((pd & (1<<PIND2))>=1  &&  (pb & (1<<PINB0))==0) ||
	   ((pd & (1<<PIND2))==0  &&  (pb & (1<<PINB0))>=1) ){
		wheel1++;
	  } 
	  else{	
		  wheel1--;
	  }
}

SIGNAL(SIG_INTERRUPT1)
{ 
	register unsigned char pd = PIND;
	register unsigned char pb = PINB;
  
  // if the signals are different is was
  // a forwad movement - else a backward movement
	if(
	   ((pd & (1<<PIND3))>=1  &&  (pb & (1<<PINB1))==0) ||
		  ((pd & (1<<PIND3))==0  &&  (pb & (1<<PINB1))>=1) 
	  ){
		wheel2++;
	  } 
	  else{	
		  wheel2--;
	  }
}

/* Interrupt Signal Handler */
//ISR(USART_RXC_vect)
SIGNAL (SIG_UART_RECV)
{
	push(interruptbuff, RXD_DATA);
}	

/* This functions transmit a packet */
byte TxPacket(byte bError, byte bParameterLength) 
{
	byte bCount,bCheckSum,bPacketLength;
	byte txbuff[ MAX_PACKELEN ];
  
	txbuff[0] = 0xff;
	txbuff[1] = 0xff;
	txbuff[2] = SELF_ID;
	txbuff[3] = bParameterLength+2; // Length(Paramter,Instruction,Checksum)
	txbuff[4] = bError;
  
	for(bCount = 0; bCount < bParameterLength; bCount++)
		txbuff[bCount+5] = param[bCount];
  
	bCheckSum = 0;
	bPacketLength = bParameterLength+4+2;
  
	for(bCount = 2; bCount < bPacketLength-1; bCount++) // except 0xff,checksum
		bCheckSum += txbuff[bCount];
  
	txbuff[bCount] = ~bCheckSum; // Writing Checksum with Bit Inversion
  
	RS485_TXD;
	for(bCount = 0; bCount < bPacketLength; bCount++)
	{
		SET_TXD_FINISH;
		TxD(txbuff[bCount]);
	}
	while(!CHECK_TXD_FINISH); // Wait until TXD Shift register empty
	RS485_RXD;
  
	return(bPacketLength);
}

void TxD(byte bTxdData)
{
	while(!TXD_READY);
	TXD_DATA = bTxdData;  
}


void callback( BioloidPacket *pkt )
{  
  // checksum test
	if( pkt->crc != 0xff ){
		TxPacket(E_CHECKSUM,0);
		return;
	}   
  
  // ping
	if( pkt->cmd == INST_PING ){
		TxPacket(NO_ERROR,0);
		return;
	}  
  
  // read
	if( pkt->cmd == INST_READ && pkt->param[0] == P_DIST ){
    
    // check the limits of the size
		if( wheel1 > WORD_MAX_SIZE ) wheel1 = WORD_MAX_SIZE;
		if( wheel1 < WORD_MIN_SIZE ) wheel1 = WORD_MIN_SIZE;
		if( wheel2 > WORD_MAX_SIZE ) wheel2 = WORD_MAX_SIZE;
		if( wheel2 < WORD_MIN_SIZE ) wheel2 = WORD_MIN_SIZE;
    
    // send a signed word for both wheels
		param[0] = lowbyte(wheel1);
		param[1] = highbyte_s(wheel1);
		param[2] = lowbyte(wheel2);
		param[3] = highbyte_s(wheel2);
		TxPacket(NO_ERROR,4);
		return;
	}
  
  // reset
	if( pkt->cmd == INST_RESET ){
		TxPacket(NO_ERROR,0);
		RESET_WHEEL;
		return;
	}
}


