#include "global.h"
#include "Controller.h"

// clock configuration
#define F_CPU 	11059200UL

/** initializes servos, states-space and LEDs.*/
void Controller::init() {

	mainMode = MAIN_MODE_STOP;
	newMainMode = MAIN_MODE_WALK_POLICY;
	states.init();

	// assign PWM-Timer to servos
	servo_x.setTimer(TIMER1B);
	servo_y.setTimer(TIMER1A);
	
	// initialize LED pins as outputs (PA0-3) and switch-pins (PA4-7) as inputs
	DDRA = (1<<PA0) | (1<<PA1) | (1<<PA2) | (1<<PA3); // LED1..4
	// enable internal pull-ups for switch-pins (PA4-7)
	PORTA |= (1<<PA4) | (1<<PA5) | (1<<PA6) | (1<<PA7);
	
	// initialize switch-pins (PB0, PB1, PB2, PB4) as inputs
	DDRB = 0x00;
	// enable internal pull-ups for switch-pins
	PORTB |= (1<<PB0) | (1<<PB1) | (1<<PB2) | (1<<PB4);
	
	// turn on all leds
	initLEDs();
	
	// enable auto-start
	autoStart = true;

	// set initial-state 
	currentX = 0x00;
	currentY = 0x00;

	// set default policy 
	setDefaultPolicy();
}

/** return the switch status !! low-active pins are converted to: On:=1, Off=0) */
unsigned char Controller::getSwitchState (unsigned char _switch) {

	bool status;

	if (_switch == SWITCH_WALK) {
		//status = (PINB & (1<<PB0)); // PCBv1
		status = (PINB & (1<<PB4)); // PCBv2
		
	} else if (_switch == SWITCH_RANDOM_STATE) {
		//status = (PINB & (1<<PB1)); // PCBv1
		status = (PINB & (1<<PB2)); // PCBv2
		
	} else if (_switch == SWITCH_INV_FB_0) {
		//status = (PINB & (1<<PB2)); // PCBv1
		status = (PINB & (1<<PB1)); // PCBv2
	
	} else if (_switch == SWITCH_ENABLE_EXPLORATION) {
		//status = (PINA & (1<<PA7)); // PCBv1
		status = (PINB & (1<<PB0)); // PCBv2
		
	} else if (_switch == SWITCH_FULL_EXP_AT_START) {
		//status = (PINA & (1<<PA6)); // PCBv1
		status = (PINA & (1<<PA4)); // PCBv2
		
	} else if (_switch == SWITCH_V24) {
		// status = (PINA & (1<<PA5)); // PCBv1
		status = (PINA & (1<<PA5)); // PCBv2
						
	} else if (_switch == SWITCH_VDBE_EGREEDY) {
		//status = (PINA & (1<<PA4)); // PCBv1
		status = (PINA & (1<<PA6)); // PCBv2

	} else if (_switch == SWITCH_RAM) {
		//status = SWITCH_RAM_VALUE; // PCBv1
		status = SWITCH_RAM_VALUE; // PCBv2

	} else if (_switch == SWITCH_GAMMA) {
		status = (PINA & (1<<PA7)); // PCBv2

	// Bei unbekanntem Kommando - wird geblinkt
	} else {
		
		register char i;
		
		for (i=0; i<100; i++) {
		
			this->powerOnLED(LED_POWER, 1);
			this->powerOnLED(LED_WALK, 0);
			this->powerOnLED(LED_LEARN, 0);
			this->powerOnLED(LED_EXPLORATION, 1);
			
			this->msDelay(250);
			
			this->powerOnLED(LED_POWER, 0);
			this->powerOnLED(LED_WALK, 1);
			this->powerOnLED(LED_LEARN, 1);
			this->powerOnLED(LED_EXPLORATION, 0);
			
			this->msDelay(250);
		}
		
		return 0;
	}
	
	return !status;
}

/** enable serial interrupt */
void Controller::enableV24Interrupt()  { 

	this->disableV24Interrupt();
 	if (this->getSwitchState(SWITCH_V24)) {
 		//UCSRB |=  (1 << RXCIE); 
 		//sei();
 	}
	//PORTC &= ~(1<<PC3);
	//UCSRB |= (1<<RXEN);
}

/** disable serial interrupt  */
void Controller::disableV24Interrupt() { 
	if (this->getSwitchState(SWITCH_V24)) {
		//UCSRB &= ~(1 << RXCIE); 
		//cli();
	}
	//PINC |= (1<<PC3);
	//PORTC |= (1<<PC3);
	//UCSRB &= ~(1<<RXEN);
}


/** delay of approx. n milli-seconds */
void Controller::msDelay (int n) {

	double maxDelay = (262.14 / (double)((11059200UL / 1000000)));

	for (int i=0; i< (n/maxDelay); i++) {
		_delay_ms (maxDelay);
	}
}

/** set default policy */
void Controller::setDefaultPolicy() {

	int _val=0;
	int _row, _column;
	STATE tmpState;

	// default grid size 5x5
	states.scalX = 5;
	states.scalY = 5;
	
	// initialize states 
	for (_row=0; _row < states.scalY; _row++) {
		for (_column=0; _column<states.scalX; _column++) {

				tmpState.value = 0.0; 

				tmpState.feedback_up = 0;
				tmpState.feedback_down = 0;
				tmpState.feedback_left = 0;
				tmpState.feedback_right = 0;
			
				states.setState (_row, _column, &tmpState);
		}
	}

	// position arm into air
	this->gotoStateWithoutSavingFeedback(0,0);
}


/** set servo to specific position without saving the reward signal */
void Controller::gotoStateWithoutSavingFeedback(unsigned char _row, unsigned char _column) {

	servo_y.setPosition(_row, states.scalY);
	servo_x.setPosition(_column, states.scalX);

	currentX = _column;
	currentY = _row;
}


/** turns a specific LED on (_status=true) of off (_status=false) */
void Controller::powerOnLED(unsigned char _ledID, bool _status) {

	if (_status == true) {
		PORTA &= ~(1<<_ledID);
	} else {
		PORTA |= (1<<_ledID);
	}
}

/** Initialization function for the LEDs */
void Controller::initLEDs() {

	// LED test (turn all LEDs on)
	this->powerOnLED(LED_POWER, true);
	this->powerOnLED(LED_WALK, true);
	this->powerOnLED(LED_LEARN, true);
	this->powerOnLED(LED_EXPLORATION, true);

	msDelay(800);

	// LED test (turn all LEDs off)
	this->powerOnLED(LED_POWER, true);
	this->powerOnLED(LED_WALK, false);
	this->powerOnLED(LED_LEARN, false);
	this->powerOnLED(LED_EXPLORATION, false);
}


void Controller::blinkLED (unsigned char led_id, unsigned int count, unsigned int delay) {

	unsigned char i;

	DDRA = 0x0F;
	PORTA |= (1<<PA4) | (1<<PA5) | (1<<PA6) | (1<<PA7);

	for (i=0; i<count; i++) {

 		if (led_id & (1<<LED_LEARN)) this->powerOnLED ( LED_LEARN, false);
 		if (led_id & (1<<LED_WALK)) this->powerOnLED ( LED_WALK, false);
 		if (led_id & (1<<LED_EXPLORATION)) this->powerOnLED ( LED_EXPLORATION, false);
 		if (led_id & (1<<LED_POWER)) this->powerOnLED ( LED_POWER, true);

		this->msDelay (delay);


 		if (led_id & (1<<LED_WALK)) this->powerOnLED ( LED_WALK, true);
 		if (led_id & (1<<LED_LEARN)) this->powerOnLED ( LED_LEARN, true);
 		if (led_id & (1<<LED_EXPLORATION)) this->powerOnLED ( LED_EXPLORATION, true);
 		if (led_id & (1<<LED_POWER)) this->powerOnLED ( LED_POWER, true);

		this->msDelay (delay);
	}
}
