#include "StateSpace.h"

/** init function. */
void StateSpace::init() {

	// use internal memory instead of FM24C256
	this->useInternalRAM = true;

	// init ram
	ram1.init();
}

/** returns a state into _state */
void StateSpace::getState (unsigned char _row, unsigned char _column, STATE *_state) {

	unsigned int _addr;
	unsigned char i;
	unsigned char* _bytes = (unsigned char*)_state;

	// compute memory address
	_addr = ((_row * scalX) + _column)*STATESIZE;

	// internal RAM
	if (this->useInternalRAM) {
		for (i=0; i<STATESIZE; i++) {
			_bytes[i] = internStateSpace[(_addr + i)];
		}

	// SPI memory
	} else {
		ram1.readBytes(_addr, (unsigned char*)&_state[0], STATESIZE);
	}
}

/** update state (_row, _column) with _state*/
void StateSpace::setState (unsigned char _row, unsigned char _column, STATE *_state) {

	unsigned int _addr;
	unsigned char i;
	unsigned char* _bytes = (unsigned char*)_state;

	// compute memory address
	_addr = ((_row * scalX) + _column)*STATESIZE;

	// internal RAM
	if (this->useInternalRAM) {
		for (i=0; i<STATESIZE; i++) {
			internStateSpace[_addr + i] = _bytes[i];
		}

	// SPI Memory
	} else {
		ram1.writeBytes(_addr, (unsigned char*)&_state[0], STATESIZE);
	}
}

/** set Value in state (_row, _column) */
void StateSpace::setValue(unsigned char _row, unsigned char _column, float _value) {

	STATE _s;
	this->getState ( _row, _column, &_s );
	_s.value = _value;
	this->setState ( _row, _column, &_s );
}

/** set reward r(s,_action)=_fb in state (_row, _column) */
void StateSpace::setFeedback(unsigned char _row, unsigned char _column, unsigned char _action, char _fb) {

	STATE _s;
	this->getState (_row, _column, &_s);

	if (_action == UP) {
		_s.feedback_up = _fb;
	} else if (_action == DOWN) {
		_s.feedback_down = _fb;
	} else if (_action == LEFT) {
		_s.feedback_left = _fb;
	} else if (_action == RIGHT) {
		_s.feedback_right = _fb;
	}

	this->setState (_row, _column, &_s);
}

/** returns the Value from state (_row, _column) */
double StateSpace::getValue(unsigned char _row, unsigned char _column) {
	STATE _s;
	this->getState (_row, _column, &_s);
	return _s.value;
}

/** returns r(s,a) from state (_row, _column) */
char StateSpace::getFeedback (unsigned char _row, unsigned char _column, unsigned char _action) {

	STATE _s;
	this->getState (_row, _column, &_s);

	if (_action == UP) {
		return _s.feedback_up;
	} else if (_action == DOWN) {
		return _s.feedback_down;
	} else if (_action == LEFT) {
		return _s.feedback_left;
	} else if (_action == RIGHT) {
		return _s.feedback_right;
	}
}


/** RAM-Test function (e.g. writes random numbers from *data to memory)  */
void StateSpace::ramtestSaveBytes(unsigned int _addr, unsigned char *data, unsigned char _length) {

	unsigned char* statesPtr = (unsigned char*)&internStateSpace[0];
	unsigned char i;

	if (this->useInternalRAM) {
		for (i=0; i<_length; i++) {
			statesPtr[_addr + i] = data[i];
		}

	} else {
		ram1.writeBytes(_addr, data, _length);
	}
}

/** RAM-Test function (e.g. reads random numbers into *data from memory)  */
void StateSpace::ramtestReadBytes(unsigned int _addr, unsigned char *data, unsigned char _length) {

	unsigned char* statesPtr = (unsigned char*)&internStateSpace[0];
	unsigned char i;

	if (this->useInternalRAM) {
		for (i=0; i<_length; i++) {
			data[i] = statesPtr[_addr + i];
		}

	} else {
		ram1.readBytes(_addr, data, _length);
	}
}

/** returns the memory-size of the state-space */
unsigned int StateSpace::getMemorySize() {

	// Interner RAM
	if (this->useInternalRAM) {
		return sizeof(internStateSpace);

	// SPI RAM
	} else {
		return 65535;
	}
}

