#include "V24Control.h"

/** 
 * global objects which are used by the interrupt function SIGNAL(SIG_UART_RECV) 
 */
extern V24Control v24;
V24COMMAND tx_command, rx_command;
volatile bool newDataAvailable = false;


/**
 * returns true if new data arrived in USART buffer, otherwise zero
 */
bool V24Control::newDataReceived() {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {

		// enable interrupt
		this->vi.ctr.enableV24Interrupt();
		
// 		if (newDataAvailable == true || (UCSRA & (1<<RXC))) {
// 			newDataAvailable = false;
// 			return true;
// 		} else {
// 			return false;
// 		}
		
		return (UCSRA & (1<<RXC));
	}

	return false;
}

/**
 * initializes the serial interface according to baud-rate MYUBRR, 1 stop bit and no parity bit
 */
void V24Control::init () {


	// initialize value iteration
	vi.init();

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		// configure ports
		PORTD |= (1<<PD0);
		PORTD &= ~(1<<PD1);
		PORTC |= (1<<PC2);
		DDRC &= ~(1<<PC2);
		DDRC |= (1<<PC3);
		PORTC &= ~(1<<PC3);
		// set baud rate
		UBRRH = (unsigned char) (MYUBRR>>8);
		UBRRL = (unsigned char) MYUBRR;
	
		// enable receiver and transmitter
		UCSRB = (1<<RXEN) | (1<<TXEN);
	
		// Frame Format: 8data, 1stop bit, noParity 
		UCSRC = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0);
	
		// clear USART buffer
		while ((UCSRA & (1<<RXC)));

		// enable receive interrupt
		vi.ctr.enableV24Interrupt();
		newDataAvailable = false;
	
		// should each received byte be send back to the sender? (for debug)
		rxtx = true;
	}
}


/**
 * transmits a byte on the serial interface (low level)
 */
void V24Control::transmit (unsigned char data) {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {

		/* wait till transmit buffer is ready */
		while (!(UCSRA & (1<<UDRE)));

		/* wait until busy-pin (from RF04) is ready */
		while (PINC & (1<<PC2));
	
		/* send data */
		UDR = data;
	
		// wait till transmission is complete
		//while (!(UCSRA & (1<<TXC)));
	}
}


/**
 * function receives one byte from the serial interface and writes it to *data (low-level)
 */
bool V24Control::receive (unsigned char *data) {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {

		vi.ctr.powerOnLED(LED_POWER, false);
		unsigned char rx;
		int errorLoop = 0;
	
		while (!(UCSRA & (1<<RXC)) && errorLoop <100) {
			errorLoop++;
			vi.ctr.msDelay(1);
		}
	
		// error check
		if (errorLoop==100) {
			vi.ctr.powerOnLED(LED_POWER, true);
			//vi.ctr.newMainMode = MAIN_MODE_STOP;
			return false;
		} else {
			// return received data
			*data = UDR;
	
			vi.ctr.powerOnLED(LED_POWER, true);
			return true;
		}	
	}
}

/**
 * transmits a command on the serial interface
 */
bool V24Control::transmitCommand (V24COMMAND *c) {

	unsigned char loopCount;

	if (vi.ctr.getSwitchState(SWITCH_V24)) {

	
		unsigned char tmp;
	
		// transmit command-byte
		transmit(c->command);
		/*if (rxtx) {
			receive(&tmp);
			if (tmp != c->command) return false;
		}*/

		// transmit command-length	
		transmit(c->dataLength);
		/*if (rxtx) {
			receive(&tmp);
			if (tmp != c->dataLength) return false;
		}*/
	
		// transmit data
		for (loopCount=0; loopCount < c->dataLength; loopCount++) {
	
			transmit(c->data[loopCount]);
			/*if (rxtx) {
				receive(&tmp);
				if (tmp != c->data[loopCount]) return false;
			}*/
		}
	
		return true;
	}
}

/**
 * receive a commando from the serial interface into *c
 */
bool V24Control::receiveCommand (V24COMMAND *c) {

	unsigned char loopCount, rData, data;
	bool sync = false;
	char i;

	if (vi.ctr.getSwitchState(SWITCH_V24)) {

		// synchronisation
		rData = !V24_COMMAND_SYNC_COMMAND;
		do {
			if (receive(&rData) && rData==V24_COMMAND_SYNC_COMMAND) {
				transmit(V24_COMMAND_SYNC_OK_COMMAND);
				sync = true;
			}
		} while (!sync);

		// receive command byte
		sync = false;
		do {
			if (receive(&c->command)) {
				transmit(c->command);
				sync = true;
			}
		} while (!sync);
		
		// receive command-length
		sync = false;
		do {
			if (receive(&c->dataLength)) {
				transmit(c->dataLength);
				sync = true;
			}
		} while (!sync);

		
		// receive command data
		for (loopCount=0; loopCount < c->dataLength; loopCount++) {
			sync = false;
			do {
				if (receive(&c->data[loopCount])) {
					transmit(c->data[loopCount]);
					sync = true;
				}
			} while (!sync);		
		}
	
		return true;
	}
}


/**
 *  process command 
 */
void V24Control::processCommand() {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		// disable all interrupts
		cli();
		this->vi.ctr.disableV24Interrupt();
	
		char i=0;
		char cmd;
		unsigned char* bytes;
		int _row, _column;
		V24COMMAND tmp_command;
			
		// turn off LEDs
		vi.ctr.powerOnLED (LED_LEARN, false);
		vi.ctr.powerOnLED (LED_WALK, false);
		vi.ctr.powerOnLED (LED_EXPLORATION, false);
	
		// disable auto start
		vi.ctr.autoStart = false;
	
		// receive command
		//if (!commandParsedByIRQ) {
			if (!this->receiveCommand ( &rx_command )) {
				vi.ctr.newMainMode = MAIN_MODE_STOP;
				vi.ctr.msDelay(1000);
				this->flushReceiveBuffer();
				this->vi.ctr.enableV24Interrupt();
				return;
			}
		//} else {
		//	commandParsedByIRQ = false;
		//}
	
		cmd = rx_command.command;
	
		/***************
		* walking mode *
		****************/
		if ( cmd == V24_COMMAND_START_WALKING ) {
	
			// go to start position (0,0)
			vi.ctr.gotoStateWithoutSavingFeedback(0, 0);
			vi.ctr.newMainMode = MAIN_MODE_WALK_POLICY;
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
		
		/*************
		* stop robot *
		**************/
		} else if ( cmd == V24_COMMAND_STOP_WALKING ) {
	
			vi.ctr.newMainMode = MAIN_MODE_STOP;
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
		
		/********************
		* send gamma to GUI *
		*********************/
		} else if ( cmd == V24_COMMAND_GET_GAMMA ) {
	
			processGetGamma();
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
		/************
		* set gamma *
		*************/
		} else if ( cmd == V24_COMMAND_SET_GAMMA ) {
	
			processSetGamma(&rx_command);
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
		/**********************
		* set grid dimensions *
		***********************/
		} else if ( cmd == V24_COMMAND_SET_GRIDSIZE ) {
	
			processSetGridsize(&rx_command);
			
		/*******************************
		* send current grid dimensions *
		********************************/
		} else if ( cmd == V24_COMMAND_GET_GRIDSIZE ) {
	
			processGetGridsize();
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
		/***************************************
		* set a state (Value+rewards)          *
		****************************************/
		} else if ( cmd == V24_COMMAND_SET_STATE ) {
	
			processSetState(&rx_command);
	
		/**************************
		* send a state to the GUI *
		***************************/
		} else if ( cmd == V24_COMMAND_GET_STATE ) {
		
			processGetState(&rx_command);
	
		/*****************************************************
		* returns the exploration/exploitation configuration *
		******************************************************/
		} else if ( cmd == V24_COMMAND_GET_EXPLORATION_CONFIG ) {
		
			processGetExplorationConfig();
	
		/**********************************************
		* returns user defined debug information *
		***********************************************/
		} else if ( cmd == V24_COMMAND_GET_DEBUG ) {
		
			processGetDebug();

		/*************************
		* RAM-Test - write bytes *
		**************************/
		} else if ( cmd == V24_COMMAND_WRITE_RAM_BYTES ) {
	
			vi.ctr.mainMode = vi.ctr.newMainMode = MAIN_MODE_STOP;
	
			// write random numbers into memory
			vi.ctr.states.ramtestSaveBytes(  
				((rx_command.data[0]*256) + rx_command.data[1]), // Address
				&(rx_command.data[2]),  // Pointer to data
				rx_command.dataLength-2 // size
	
			);
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
		/************************
		* RAM-Test - read bytes *
		*************************/
		} else if ( cmd == V24_COMMAND_READ_RAM_BYTES ) {
	
			// stop main program
			vi.ctr.mainMode = vi.ctr.newMainMode = MAIN_MODE_STOP;
	
			// create commando structure
			tmp_command.command = V24_COMMAND_WRITE_RAM_BYTES;
			tmp_command.dataLength = rx_command.data[2]+2; //
			tmp_command.data[0] = rx_command.data[0];
			tmp_command.data[1] = rx_command.data[1];
	
			// read data from RAM
			vi.ctr.states.ramtestReadBytes(  
				((rx_command.data[0]*256) + rx_command.data[1]), // address
				&(tmp_command.data[2]),	 // data pointer
				rx_command.data[2]	 // size
			);
	
			// send back to GUI
			if (!v24.transmitCommand (&tmp_command)) return;
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
		/***********************************
		* explore all rewards		   *
		************************************/
		} else if ( cmd == V24_COMMAND_EXPLORE_FEEDBACK ) {
	
			vi.ctr.newMainMode = MAIN_MODE_EXPLORE_FEEDBACK;
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
		/***********************************
		* learning only mode 		   *
		************************************/
		} else if ( cmd == V24_COMMAND_START_LEARNING ) {
	
			vi.ctr.newMainMode = MAIN_MODE_LEARN;
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;

		/**********************************
		* returns the distance traveled   *
		***********************************/
		} else if (cmd == V24_COMMAND_GET_DISTANCE ) {

			signed long int _distance = vi.getDistanceCovered();
			unsigned char* bytes = (unsigned char*)&_distance;

			tx_command.command = V24_COMMAND_SET_DISTANCE;
			tx_command.dataLength = 0x04;
			tx_command.data[0] = bytes[0];
			tx_command.data[1] = bytes[1];
			tx_command.data[2] = bytes[2];
			tx_command.data[3] = bytes[3];

			if (!v24.transmitCommand (&tx_command)) return; 
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;

		/**************************
 		 * reset distance counter *
 		 **************************/
		} else if (cmd == V24_COMMAND_RESET_DISTANCE_COUNTER ) {

			vi.resetDistanceCounter();
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;

		/***************************
		 * set exploration config  *
		 ***************************/
		} else if (cmd == V24_COMMAND_SET_EXPLORATION_CONFIG ) {

			unsigned char* bytes;
			register char i;
		
			// exploration parameter
			bytes = (unsigned char*)&vi.expParameter;
			for (i=0; i < sizeof(vi.expParameter); i++) {
				bytes[i] = rx_command.data[i];
			}

			// exploration mode			
			bytes = (unsigned char*)&vi.expMode;
			bytes[0] = rx_command.data[sizeof(vi.expParameter)];

			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;

		/***********************************
		* send current possition to GUI    *
		***********************************/
		} else if (cmd == V24_COMMAND_GET_CURRENTPOSITION_MODE_ITERATIONS ) {
	
			unsigned long int _itCount = vi.getIterationCount();
			unsigned char* bytes = (unsigned char*)&_itCount;
			signed char wert;
			
			tx_command.command = V24_COMMAND_SET_CURRENTPOSITION_MODE_ITERATIONS;
			tx_command.dataLength = 0x09;
			tx_command.data[0] = vi.ctr.currentY;
			tx_command.data[1] = vi.ctr.currentX;
			tx_command.data[2] = (unsigned char)vi.ctr.mainMode;
			tx_command.data[3] = bytes[0];
			tx_command.data[4] = bytes[1];
			tx_command.data[5] = bytes[2];
			tx_command.data[6] = bytes[3];
			wert = 127;
			tx_command.data[7] = *(unsigned char*)&wert;
			wert = -127;
			tx_command.data[8] = *(unsigned char*)&wert;
			
	
			if (!v24.transmitCommand (&tx_command)) return; 
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
		/*********************************************************
		* receive states (values+rewards) and gamma from the GUI *
		**********************************************************/
		} else if (cmd == V24_COMMAND_SET_ALL_STATES_AND_GAMMA ) {
		
			// set gamma
			if (!this->receiveCommand(&tmp_command)) return;
			processSetGamma(&tmp_command);
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
			// set grid size
			if (!this->receiveCommand(&tmp_command)) return;
			processSetGridsize(&tmp_command);
	
			for (_row=0; _row < vi.ctr.states.scalY; _row++) {
				for (_column=0; _column < vi.ctr.states.scalX; _column++) {
	
					// receive state
					if (!this->receiveCommand (&tmp_command)) return;
					processSetState(&tmp_command);
				}
			}		
		
		/********************************************************
		* send all states (Values+Rewards) and gamma to the GUI *
		*********************************************************/
		} else if (cmd == V24_COMMAND_GET_ALL_STATES_AND_GAMMA ) {
	
			// send gamma
			if (!this->receiveCommand(&tmp_command)) return;
			processGetGamma();
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
			// send grid size
			if (!this->receiveCommand(&tmp_command)) return;
			processGetGridsize();
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;
	
			for (_row=0; _row < vi.ctr.states.scalY; _row++) {
				for (_column=0; _column < vi.ctr.states.scalX; _column++) {
	
					// receive getState Command
					if (!this->receiveCommand (&tmp_command)) return;

					// send state
					processGetState(&tmp_command);
				}
			}		

		// perform next action and return reward to the GUI
		} else if (cmd == V24_COMMAND_WALK_AND_RETURN ) {


			tx_command.command = V24_COMMAND_WALK_AND_RETURN;
			tx_command.dataLength = 1;
			tx_command.data[0] = vi.moveAndSaveFeedback(rx_command.data[0], 0);

			if (!v24.transmitCommand (&tx_command)) return; 
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return;

		/*******************************************************************************
		* return "UNKNOWN COMMAND" in case of an unknown command
		*******************************************************************************/
		} else {
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_ERROR, V24_ERRORCODE_UNKNOWN_COMMAND)) return;
		}
	
	
		// enable serial interrupt
		this->vi.ctr.enableV24Interrupt();
	}
}

/**
 * debug: ..
 */
bool V24Control::transmitReceiveCommandProcessed(unsigned char _code, unsigned char _error) {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
// 		V24COMMAND ok_command;
// 		ok_command.command = _code;
// 	
// 		// Im Fehlerfall, den Fehlercode anhängen
// 		if (_error > 0) {
// 			ok_command.dataLength = 0x01;
// 			ok_command.data[0] = _error;
// 	
// 		// Ansonsten gibt es keinen Datenteil
// 		} else {
// 			ok_command.dataLength = 0x00;
// 		}
// 	
// 		// Bestätigung senden
// 		if (!v24.transmitCommand(&ok_command)) return false;
// 	
	// 	return true;
	
	}
	return true;
}

/**
 * reset the serial buffer 
 */
void V24Control::flushReceiveBuffer() {
	if (vi.ctr.getSwitchState(SWITCH_V24)) {
		unsigned char dummy;
		while (UCSRA & (1<<RXC)) dummy = UDR;
	}
}


/** 
 * returns exploration parameter
 */
bool V24Control::processGetExplorationConfig() {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		V24COMMAND tx_command;
		unsigned char* bytes;
		register char i;
	
		tx_command.command = V24_COMMAND_SET_EXPLORATION_CONFIG;
		tx_command.dataLength = 5;

		// Exploration Parameter
		bytes = (unsigned char*)&(vi.expParameter);
		for (i=0; i < sizeof(vi.expParameter); i++) {
			tx_command.data[i] = bytes[i];
		}

		// Exploration Mode
		tx_command.data[4] = vi.expMode;
		
	
		if (!transmitCommand (&tx_command)) return false;
		return true;
	}
}


/** 
 * send gamma to GUI
 */
bool V24Control::processGetGamma() {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		V24COMMAND tx_command;
		unsigned char* bytes;
		register char i;
	
		tx_command.command = V24_COMMAND_SET_GAMMA;
		tx_command.dataLength = 4;
		bytes = (unsigned char*)&(vi.gamma);
	
		for (i=0; i < tx_command.dataLength; i++) {
			tx_command.data[i] = bytes[i];
		}
	
		if (!transmitCommand (&tx_command)) return false;
		return true;
	}
}

/**
 * set gamma in Value-Iteration
 */
void V24Control::processSetGamma(V24COMMAND *cmd) {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		unsigned char* bytes;
		register char i;
	
		bytes = (unsigned char*)&vi.gamma;
	
		for (i=0; i < 4; i++) {
			bytes[i] = cmd->data[i];
		}
	}
}

/** 
 * returns user defined debug information
 */
bool V24Control::processGetDebug() {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		V24COMMAND tx_command;
	
		tx_command.command = V24_COMMAND_GET_DEBUG;
		tx_command.dataLength = 10;

		register char i;

		unsigned char* bytes = (unsigned char*)&(vi.currentExpProbability);
		for (i=0; i < tx_command.dataLength; i++) {
			tx_command.data[i] = bytes[i];
		}

		for (i=4; i<tx_command.dataLength; i++) {
			tx_command.data[i] = rand()%(NUM_ACTIONS);
			srand (rand()+1);
		}
	
		if (!transmitCommand (&tx_command)) return false;
		return true;
	}
}


/** 
 * send grid-size to GUI
 */
bool V24Control::processGetGridsize() {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		V24COMMAND tx_command;
	
		tx_command.command = V24_COMMAND_SET_GRIDSIZE;
		tx_command.dataLength = 2;
		tx_command.data[0] = vi.ctr.states.scalY;
		tx_command.data[1] = vi.ctr.states.scalX;
	
		if (!transmitCommand (&tx_command)) return false;
		return true;
	}
}

/** 
 * set local grid-size
 */
bool V24Control::processSetGridsize(V24COMMAND *cmd) { 

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		unsigned char _scalX, _scalY;
	
		_scalY = cmd->data[0];
		_scalX = cmd->data[1];
	
		// check for maximum statespace size
		if ((_scalX*_scalY) > (MAX_STATES)) {
			if (!transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_ERROR, V24_ERRORCODE_GRIDSIZE_TOO_LARGE)) return false;
			return false;

		} else {
			vi.ctr.states.scalY = _scalY;
			vi.ctr.states.scalX = _scalX;
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return false;
			return true;
		}
	}
}

/**
 * send a specific state to the GUI
 */
bool V24Control::processGetState(V24COMMAND *cmd) {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		V24COMMAND tx_command;
		unsigned char _row = cmd->data[0];
		unsigned char _column = cmd->data[1];
		
		if (_row >= vi.ctr.states.scalY || _column >= vi.ctr.states.scalX) {
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_ERROR, V24_ERRORCODE_GRIDSIZE_TOO_LARGE)) return false;
			return false;
		} else {
	
			// get state
			vi.ctr.states.getState(_row, _column, &tmpState);
			unsigned char* bytes = (unsigned char*)&tmpState;
		
			tx_command.command = V24_COMMAND_SET_STATE;
			tx_command.dataLength = STATESIZE+2;
			tx_command.data[0] = _row;
			tx_command.data[1] = _column;
		
			for (unsigned char i=0; i < STATESIZE; i++) {
				tx_command.data[2 + i] = bytes[i];
			}
	
			if (!v24.transmitCommand (&tx_command)) return false;
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return false;
			return true;
		}
	}
}

/** 
 * set received state (value+rewards) 
 */
bool V24Control::processSetState(V24COMMAND *cmd) {

	if (vi.ctr.getSwitchState(SWITCH_V24)) {
	
		unsigned char _row = cmd->data[0];
		unsigned char _column = cmd->data[1];
	
		unsigned char* bytes = (unsigned char*)&tmpState;
	
		// check, if state exists in grid
		if (_row >= vi.ctr.states.scalY || _column >= vi.ctr.states.scalX) {
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_ERROR, V24_ERRORCODE_GRIDSIZE_TOO_LARGE)) return false;
			return false;
		} else {
		
			for (unsigned char i=0; i<STATESIZE; i++) {
				bytes[i] = cmd->data[i + 2];
			}
	
			// save state to memory
			vi.ctr.states.setState (_row, _column, &tmpState);
			if (!v24.transmitReceiveCommandProcessed (V24_COMMAND_COMMAND_OK, 0)) return false;
			return true;
		}
	}
}


/** Interrupt-Service-Routine (ISR) for command reception. */
//SIGNAL (SIG_UART_RECV) {

// ISR (USART_RXC_vect) {
//  	newDataAvailable = true;
// }
//void SIG_UART_RECV (void) __attribute__ ((signal));
//void SIG_UART_RECV (void) {

// 	if (v24.vi.ctr.getSwitchState(SWITCH_V24)) {
// 	
// 		// disable interrups
// 		cli();
// 		v24.vi.ctr.disableV24Interrupt();
// 	
// 		// signal that a new command arrived
// 		newDataAvailable = true;
// 		commandParsedByIRQ = true;	
// 	
// 		// switch off LEDs
// 		v24.vi.ctr.powerOnLED (LED_LEARN, false);
// 		v24.vi.ctr.powerOnLED (LED_WALK, false);
// 		v24.vi.ctr.powerOnLED (LED_EXPLORATION, false);
// 	
// 		// receive command
// 		if(!v24.receiveCommand ( &rx_command )) return;
// 	
// 		// send immediate answer for some commands
// 		if (rx_command.command == V24_COMMAND_GET_CURRENTPOSITION_MODE_ITERATIONS ||
// 		rx_command.command == V24_COMMAND_GET_GRIDSIZE ||
// 		rx_command.command == V24_COMMAND_GET_GAMMA ) {
// 	
// 			// signal that command has already been processed
// 			newDataAvailable = false;
// 			
// 			// process command
// 			v24.processCommand();
// 	
// 			// enable interrupts
// 			v24.vi.ctr.enableV24Interrupt();
// 			sei();
// 		}
// 	}
//}


