#include "ValueIteration.h"
//#include <avr/io.h>
 

/**
 * Init function: initializes incremental encoder, gamma, PRNG and controller object
 */
void ValueIteration::init() {

	fb.init();
	ctr.init();

	// seed for PRNG
	srand (get_seed() + 1103515245);


	// set initial gamma
	if (ctr.getSwitchState(SWITCH_GAMMA)) {
		this->gamma = 0.98;
	} else {
		this->gamma = 0.9;
	}

	// set initial exploration mode
	this->expMode = ctr.getSwitchState(SWITCH_VDBE_EGREEDY);

	this->iteration = 0;
	this->currentExpProbability = 1.0;

	if (this->expMode == EXPLORATION_EGREEDY) {
		this->expParameter = 0.075;
	} else if (this->expMode == EXPLORATION_VALUE_BASED) {
		this->expParameter = 1;
	}

	oldValue = newValue = 0.0;
	this->distanceCovered = 0;
}

/**
 * returns the amount of value-iterations performed so long
 *
 */
unsigned long int ValueIteration::getIterationCount() {
	return this->iteration;
}

/**
 * returns the traveled distance of the crawling robot
 */
signed long int ValueIteration::getDistanceCovered() {
	return this->distanceCovered;
}

/**
 * resets the distance counter
 */
void ValueIteration::resetDistanceCounter() {
	this->distanceCovered = 0;
}

/**
 * Returns the greedy action in state (_row, _column)
 */
unsigned char ValueIteration::getGreedyAction(unsigned char _row, unsigned char _column) {

	float value, maximum = -100;
	unsigned short multipleBestActions = 0;
	unsigned char _action, bestAction;
	bool haveAction = false;


	// 1) calculate values for all valid actions
	for(_action=0; _action <NUM_ACTIONS; _action++) {

		if (isValidAction(_row, _column, _action)) {

			// calculate value
			value = evaluateAction (_row, _column, _action);
	
			// check for maximum
			if(value > maximum) {
	
				maximum = value;
				haveAction = true;
				
				// memory the best action
				bestAction = _action;
	
				multipleBestActions = 0;
	
			} else  if (fabs(value - maximum) < 0.001) {
				multipleBestActions++;
			}
		}
	}

	// if there are multiple best actions, chose a random one 
	if (multipleBestActions) {

		unsigned char found=0;
		unsigned char randAction = rand()%(multipleBestActions+1);
	
		for (_action=0; _action < NUM_ACTIONS; _action++) {
			if (isValidAction(_row, _column, _action)) {
				if (fabs((value=evaluateAction(_row, _column, _action)) - maximum) < 0.001) {
					if (found < randAction) {
						found++;
					} else {
						bestAction = _action;
						_action = NUM_ACTIONS; // terminates the for-loop
					}
				}
			}
		}
	}
	

	// 2) if there was no best action found, take random action
	if (!haveAction) {	
		bestAction= this->getRandomAction();
		ctr.blinkLED((1<<LED_EXPLORATION), 100, 50);
	}

	return bestAction;
}


/**
 * this functino performs the value-iteration algorithm for all states
 */
void ValueIteration::valueIterateAllStates() {

	float value, maximum = -100;
	unsigned char bestAction, action, _row, _column;
	unsigned short multipleBestActions = 0;
	bool haveAction = false;
	STATE _tmpState;
	
	// increment iteration number
	this->iteration++;

	// switch on LEARN LED
	ctr.powerOnLED(LED_LEARN, true);
	
	// for each state s in S: 
	for(_row=0; _row<ctr.states.scalY; _row++) {
		for(_column=0; _column<ctr.states.scalX; _column++) {

			haveAction = false;
			maximum = -100;
			multipleBestActions=0;

			// 1) for each valid successor state
			for(action=0; action<NUM_ACTIONS; action++) {

				if (isValidAction(_row, _column, action)) {

					value = evaluateAction (_row, _column, action);

					// memorize maximum value
					if(value > maximum) {

						maximum = value;
						haveAction = true;

						// update maximum value in state-space
						ctr.states.setValue(_row, _column, value);
					}	
				}
			}
		}
	}

	// switch off LEARN LED
	ctr.powerOnLED(LED_LEARN, false);
}


/**
 * compute the successor-state value for action a in state (_row, _column)
 *   v' = feedback(a) + gamma*v(a)
 */
float ValueIteration::evaluateAction(unsigned char _row, unsigned char _column, unsigned char _action) {

	float neighbourValue, newValue;
	int actionFeedback;
	
	unsigned char _newRow = _row;
	unsigned char _newColumn = _column;

	if (_action == UP)   _newRow--;
	if (_action == DOWN) _newRow++;
	if (_action == RIGHT) _newColumn++;
	if (_action == LEFT)  _newColumn--;

	// get reward for action a in s
	actionFeedback = ctr.states.getFeedback(_row, _column, _action);

	// get value of successor state
	neighbourValue = ctr.states.getValue(_newRow, _newColumn);

	// compute new value
	newValue = (gamma * (float)neighbourValue) + (float)actionFeedback;

	return newValue;
}


/**
 * perform a complete exploration of the environment. (all state-transitions)
 */
void ValueIteration::exploreAllFeedbacks(bool wait5Seconds) {

	bool direction;

	unsigned char i, j;
	unsigned char steps = 0;

	// switch on EXPLORATION LED
	ctr.powerOnLED(LED_EXPLORATION, true);

	// go to state (0,0)
	ctr.gotoStateWithoutSavingFeedback(0,0);
	direction = true;

	// delay five seconds
	if (wait5Seconds) ctr.msDelay(5000);

	// explore horizontal until (scalX, scalY)
	while (steps < (ctr.states.scalY*(ctr.states.scalX-1))) {
		
		// go right
		if (direction) {
	
			moveAndSaveFeedback(RIGHT, false);

			// go down at the right end
			if (!isValidAction(ctr.currentY, ctr.currentX, RIGHT)) {
				direction = false;
				if (isValidAction(ctr.currentY, ctr.currentX, DOWN)) {
					moveAndSaveFeedback(DOWN, false);
				}
			}

		// go left
		} else {
	
			moveAndSaveFeedback(LEFT, false);

			// go down on the left end
			if (!isValidAction(ctr.currentY, ctr.currentX, LEFT)) {
				direction = true;
				if (isValidAction(ctr.currentY, ctr.currentX, DOWN)) {
					moveAndSaveFeedback(DOWN, false);
				}
			}
		}
		steps++;
	}

	// go back horizontally from (scalX, scalY) to (0,0);
	steps=0;
	while (steps < (ctr.states.scalY*(ctr.states.scalX-1))) {
		
		// go right 
		if (direction) {
	
			moveAndSaveFeedback(RIGHT, false);
	
			// go up on the right end
			if (!isValidAction(ctr.currentY, ctr.currentX, RIGHT)) {
				direction = false;
				if (isValidAction(ctr.currentY, ctr.currentX, UP)){ 
					moveAndSaveFeedback(UP, false);
				}
			}

		// go left
		} else {
	
			moveAndSaveFeedback(LEFT, false);
			
			// go up on the left end
			if (!isValidAction(ctr.currentY, ctr.currentX, LEFT)) {
				direction = true;
				if (isValidAction(ctr.currentY, ctr.currentX, UP)) {
					moveAndSaveFeedback(UP, false);
				}
			}
		}
		steps++;

	}

	// explore vertically
	direction = true;
	steps = 0;
	while (steps < (ctr.states.scalY*(ctr.states.scalX-1))) {
		
		// go down 
		if (direction) {
	
			moveAndSaveFeedback(DOWN, false);

			// go left on the lower end
			if (!isValidAction(ctr.currentY, ctr.currentX, DOWN)) {
				direction = false;
				if (isValidAction(ctr.currentY, ctr.currentX, RIGHT)) {
					moveAndSaveFeedback(RIGHT, false);
				}
			}

		// go up
		} else {
	
			moveAndSaveFeedback(UP, false);

			// go right on the upper end
			if (!isValidAction(ctr.currentY, ctr.currentX, UP)) {
				direction = true;
				if (isValidAction(ctr.currentY, ctr.currentX, RIGHT)) {
					moveAndSaveFeedback(RIGHT, false);
				}
			}
		}
		steps++;
	}

	// go vertically backwards to (0,0);
	steps=0;
	while (steps < (ctr.states.scalY*(ctr.states.scalX-1))) {
		
		// go down
		if (direction) {
	
			moveAndSaveFeedback(DOWN, false);
	
			// go left on the lower end
			if (!isValidAction(ctr.currentY, ctr.currentX, DOWN)) {
				direction = false;
				if (isValidAction(ctr.currentY, ctr.currentX, LEFT)) {
					moveAndSaveFeedback(LEFT, false);
				}
			}

		// go left
		} else {
	
			moveAndSaveFeedback(UP, false);
			
			// go left on the upper end
			if (!isValidAction(ctr.currentY, ctr.currentX, UP)) {
				direction = true;
				if (isValidAction(ctr.currentY, ctr.currentX, LEFT)) {
					moveAndSaveFeedback(LEFT, false);
				}
			}
		}
		steps++;

	}

	// turn off FEEDBACK LED
	ctr.powerOnLED(LED_EXPLORATION, false);
}

/**
 * returns a valid random action 
**/
unsigned char ValueIteration::getRandomAction () {

	bool found = false;
	unsigned char randomAction;
	static unsigned char increment = 0;

	do {
		randomAction = (rand()+increment)%NUM_ACTIONS;
		if (isValidAction(ctr.currentY, ctr.currentX, randomAction)) { 
			found = true;
		}

		increment++;

	} while (!found);

	return randomAction;
}

/**
 * Go to a random state and update reward table
**/
void ValueIteration::gotoRandomState() {

	unsigned char newX = rand()%ctr.states.scalX;
	unsigned char newY = rand()%ctr.states.scalY;

	unsigned char oldX = ctr.currentX;
	unsigned char oldY = ctr.currentY;


	// position new state on Y-axis
	while (oldY != newY) {
		if (oldY < newY) {
			oldY++;
			this->moveAndSaveFeedback(DOWN, 0);
		}
		else {
			oldY--;
			this->moveAndSaveFeedback(UP, 0);
		}
	}
	// position new state on X-axis
	while (oldX != newX) {
		if (oldX < newX) {
			oldX++;
			this->moveAndSaveFeedback(RIGHT, 0);
		}
		else {
			oldX--;
			this->moveAndSaveFeedback(LEFT, 0);
		}
	}
}

/** returns the next Action with respect to the exploration parameters */
bool ValueIteration::doExploration() {
	
	float explorationProbability = 1;
	float dV;

	if (this->expMode == EXPLORATION_VALUE_BASED) {
		
		// calculate dV
		dV = fabs ((oldValue-newValue) / (fabs(oldValue)+fabs(newValue)));
		explorationProbability = pow(dV, 1.0/this->expParameter);

	} else if (this->expMode == EXPLORATION_EGREEDY) {
		explorationProbability = this->expParameter;
	}

	this->currentExpProbability = explorationProbability;

	// return whether we should explore or exploit
	if ((rand()%100) > (explorationProbability*100)) {
		return false; // exploitation
	} else {
		return true; // exploration
	}
}

/**
 * perform _action in currentState. _activateRandomMoving defines if exploration/exploitation should be balanced by eGreedy or VDBE. Otherwise, pure greedy exploitation is performed.
 */
char ValueIteration::moveAndSaveFeedback (unsigned char _action, bool _activateRandomMoving) {

	char _feedback;
	unsigned char oldX = ctr.currentX;
	unsigned char oldY = ctr.currentY;
	unsigned char randomAction=0;

	unsigned char nextAction = _action;
	
	STATE curState;
	ctr.states.getState(ctr.currentY, ctr.currentX, &curState);

	ctr.powerOnLED(LED_WALK, true);
	
	// check for reward-signal invertion
	if (ctr.getSwitchState(SWITCH_INV_FB_0)==0) {
		fb.invertSensor(FEEDBACK_SENSOR_0, true);
	} else {
		fb.invertSensor(FEEDBACK_SENSOR_0, false);
	}


	// check whether we should explore
	if (_activateRandomMoving && doExploration()) {

		ctr.powerOnLED(LED_EXPLORATION, true);
		ctr.powerOnLED(LED_WALK, true);

		if (ctr.getSwitchState(SWITCH_RANDOM_STATE)==0) {
			nextAction = getRandomAction();
			this->moveAndSaveFeedback(nextAction, 0);
		} else {
			this->gotoRandomState();
		}

		ctr.powerOnLED(LED_EXPLORATION, false);
		ctr.powerOnLED(LED_WALK, false);
		return 0;
	}

	
	//  if _action is invalid, take a (valid) random action
	while (!isValidAction(ctr.currentY, ctr.currentX, nextAction)) {
		nextAction = getRandomAction();
		//ctr.blinkLED((1<<LED_WALK), 1, 50);
	}


	// compute new state in grid
	if (nextAction == UP) {
		ctr.currentY--;

	} else if (nextAction == DOWN) {
		ctr.currentY++;

	} else if (nextAction == LEFT) {
		ctr.currentX--;

	} else if (nextAction == RIGHT) {
		ctr.currentX++;

	} else {
		ctr.blinkLED((1<<LED_WALK)|(1<<LED_POWER), 100, 250);
	}

	// enable reward sensor and disable serial interface
	ctr.disableV24Interrupt();
	fb.enableSensor();

	// go to new state
	if (oldX != ctr.currentX) ctr.servo_x.setPosition(ctr.currentX, ctr.states.scalX);
	if (oldY != ctr.currentY) ctr.servo_y.setPosition(ctr.currentY, ctr.states.scalY);

	// short delay
	if (nextAction == UP  || nextAction == DOWN) { 
		ctr.msDelay (
			(ctr.getSwitchState(SWITCH_RAM) ?  
				(((int)(DELAY_TIME*0.66)) / (int)ctr.states.scalY) : 
				(((int)DELAY_TIME) / (int)ctr.states.scalY))
		);
	}
	else if (nextAction == LEFT || nextAction == RIGHT) { 
		ctr.msDelay (
			(ctr.getSwitchState(SWITCH_RAM) ?  
				(((int)(DELAY_TIME*0.66)) / (int)ctr.states.scalX) : 
				(((int)DELAY_TIME) / (int)ctr.states.scalX))
		);
	}

	// receive reward, deactive reward sensor and enable serial interface
	// Feedback-Sensor wieder deaktivieren (und seriellen interrupt wieder aktivieren)
	fb.disableSensor();
	ctr.enableV24Interrupt();
	_feedback = fb.getFeedback();
	
	// reset feed (with noise from the environment :-) 
	srand (rand()+1+_feedback);

	this->distanceCovered += _feedback;
	ctr.states.setFeedback(oldY, oldX, nextAction, _feedback);

	// turn of EXPLORATION and WALK LEDs
	ctr.powerOnLED(LED_EXPLORATION, false);
	ctr.powerOnLED(LED_WALK, false);

	return _feedback;
}


/**
 * this function returns whether _action is a valid function in state (_row, _column) or not
 */ 
bool ValueIteration::isValidAction(unsigned char _row, unsigned char _column, unsigned char _action) {

	if (_action != UP && _action != DOWN && _action != LEFT && _action != RIGHT) return false;
	else if (_action == UP   && _row == 0) return false;
	else if (_action == DOWN && _row >= (ctr.states.scalY-1)) return false;
	else if (_action == LEFT  && _column == 0) return false;
	else if (_action == RIGHT && _column >= (ctr.states.scalX-1)) return false;
	
	return true;
}

unsigned short ValueIteration::get_seed()
{
   unsigned short seed = 0;
   unsigned short *p = (unsigned short*) (RAMEND+1);
   extern unsigned short __heap_start;

   while (p >= &__heap_start + 1)
      seed ^= * (--p);

   return seed;
}

