/*
*  C Implementation: Environment
*
* Description: Implements the environment where actions can be performed
*
*
* Author: Michel Tokic <michel@tokic.com>, (C) 2009
*
* Copyright: See COPYING file that comes with this distribution
*
*/
#include "Environment.h"


/**
 * initializes the environment 
 */
Environment::Environment(ENV_POS* d) {

	// set the discretization
	this->discretization.x = d->x;
	this->discretization.y = d->y;

	// initialize the wheel encoder
	//r = &FeedbackSensor();

	// initialize servos
	BioloidServo::scanBioloidBus();

	servo_x = new BioloidServo(6, 180, 450); // area of operations from 200-500
	servo_y = new BioloidServo(2, 320, 700); // area of operations from 450-600
	//servo_x = new BioloidServo(SERVO_X, 180, 450); // area of operations from 200-500
	//servo_y = new BioloidServo(SERVO_Y, 320, 700); // area of operations from 450-600


	// position servos to (0,0)
	this->currentPos.x = this->currentPos.y = 0;
	servo_x->setPosition(0, this->discretization.x);
	servo_y->setPosition(0, this->discretization.y);
	

	// instantiate FeedbackSensor
	r = new FeedbackSensor();
}


Environment::~Environment() {
	delete r;
}

/**
 * performs an action within the environment and returns the successorState + reward
 */
void Environment::doAction (unsigned int action, unsigned int delayTime, ACTION_RETURN* actionReturn) { 

	ENV_POS newPos;
	this->getCurrentState(&newPos);

	BoardController::powerOnLED((1<<LED_WALK), true);

	// initialize reward
	actionReturn->reward = 0;

	// Enable Bioloid Bus
	BoardController::enableSerialBioloidInterface();

	// if action is valid => select action
	if (this->isValidAction(&this->currentPos, action)) {

		// compute successor state
		if (action == UP) {
			newPos.y--;
		} else if (action == DOWN) { 
			newPos.y++;
		} else if (action == LEFT) {
			newPos.x--;
		} else if (action == RIGHT) {
			newPos.x++;
		}

		// perform action
		if (newPos.x != this->currentPos.x || newPos.y != this->currentPos.y) {

				r->enableSensor();
				// enable bioloid bus
				servo_x->setPosition(newPos.x, this->discretization.x);
				servo_y->setPosition(newPos.y, this->discretization.y);

				BoardController::msDelay(delayTime);

				// retransmit for the case if an error occured the first time
				servo_x->setPosition(newPos.x, this->discretization.x);
				servo_y->setPosition(newPos.y, this->discretization.y);

				r->disableSensor();
				
				// save new state information
				this->currentPos.x = newPos.x;
				this->currentPos.y = newPos.y;
	
				// assemble actionReturn
				actionReturn->reward = r->getFeedback();

		}
	} else {
		BoardController::blinkLED((1<<LED_WALK), 4, 300);
	}


	// return current state state
	actionReturn->state.x = this->currentPos.x;
	actionReturn->state.y = this->currentPos.y;

	BoardController::powerOnLED((1<<LED_WALK), false);

	// Disable Bioloid Bus and enable V24
	BoardController::disableSerialInterfaces();


}

/** returns the current state */
void Environment::getCurrentState (ENV_POS* _pos) {
	_pos->x = this->currentPos.x;
	_pos->y = this->currentPos.y;
}

/** set's the environment discretization */
void Environment::setDiscretization(ENV_POS* _discretization) {
	this->discretization.x = _discretization->x;
	this->discretization.y = _discretization->y;
}

/** set's the environment discretization */
void Environment::getDiscretization(ENV_POS* _discretization) {
	_discretization->x = this->discretization.x;
	_discretization->y = this->discretization.y;
}

/** beam robot to state (pos_x, pos_y) without returning the reward */
void Environment::beamRobotToState(ENV_POS* newPos) {

	//BoardController::enableSerialBioloidInterface();

	if (newPos->x >= 0 && newPos->x < discretization.x) {
		servo_x->setPosition(newPos->x, discretization.x);
		this->currentPos.x = newPos->x;
	} else {
		newPos->x = this->currentPos.x;
	}

	if (newPos->y >= 0 && newPos->x < discretization.y) {
		servo_y->setPosition(newPos->y, discretization.y);
		this->currentPos.y = newPos->y;
	} else {
		newPos->y = this->currentPos.y;
	}
}

/**
 * this function returns whether _action is a valid function in state (_row, _column) or not
 */ 
bool Environment::isValidAction(ENV_POS* pos, unsigned char action) {

	// check for valid action
	if (action != UP && action != DOWN && action != LEFT && action != RIGHT) return false;
	
	// check if pos is within statespace range
	if (pos->x < 0 || pos->x >= this->discretization.x ||
	    pos->y < 0 || pos->y >= this->discretization.y)
		return false;

	// check if successorstate is within the statespace
	if      (action == UP    && pos->y == 0) return false;
	else if (action == DOWN  && pos->y >= (this->discretization.y-1)) return false;
	else if (action == LEFT  && pos->x == 0) return false;
	else if (action == RIGHT && pos->x >= (this->discretization.x-1)) return false;
	
	return true;
}

