#include "ValueIteration.h"
//#include <avr/io.h>


/**
 * initializes ValueIteration with the environment
 */
ValueIteration::ValueIteration (ENV_POS* d) {


//#ifdef TESTMODE
//#warning VI: skipped initalization of the StateSpace
//#else
	// create new MxN environment
	this->env = new Environment(d);

	// seed for PRNG
	srand (get_seed() + 1103515245);

	// set initial gamma
	if (BoardController::getSwitchState(SWITCH_GAMMA)) {
		this->gamma = 0.98;
	} else {
		this->gamma = 0.9;
	}

	// set initial exploration mode
	this->expMode = BoardController::getSwitchState(SWITCH_VDBE_EGREEDY);

	this->iteration = 0;
	this->currentExpProbability = 1.0;

	if (this->expMode == EXPLORATION_EGREEDY) {
		this->expParameter = 0.075;
	} else if (this->expMode == EXPLORATION_VALUE_BASED) {
		this->expParameter = 1;
	}

	oldValue = newValue = 0.0;
	this->distanceCovered = 0;

	// initialize StateSpace
	this->states = new StateSpace(env);//(StateSpace*) malloc( sizeof(StateSpace) );
	//this->states->StateSpace::StateSpace(env);
//#endif
}

/** destructor */
ValueIteration::~ValueIteration() {
	delete this->states;
	delete this->env;
}

/**
 * returns the amount of value-iterations performed so long
 *
 */
unsigned long int ValueIteration::getIterationCount() {
	return this->iteration;
}

/**
 * returns the traveled distance of the crawling robot
 */
signed long int ValueIteration::getDistanceCovered() {
	return this->distanceCovered;
}

/**
 * resets the distance counter
 */
void ValueIteration::resetDistanceCounter() {
	this->distanceCovered = 0;
}

/**
 * Returns the greedy action in state (_row, _column)
 */
unsigned char ValueIteration::getGreedyAction(ENV_POS* pos) {

	float value, maximum = -100;
	unsigned short multipleBestActions = 0;
	unsigned char _action, bestAction;
	bool haveAction = false;


	// 1) calculate values for all valid actions
	for(_action=0; _action <NUM_ACTIONS; _action++) {

		if (this->env->isValidAction(pos, _action)) {

			// calculate value
			value = evaluateAction (pos, _action);
	
			// check for maximum
			if(value > maximum) {
	
				maximum = value;
				haveAction = true;
				
				// memory the best action
				bestAction = _action;
	
				multipleBestActions = 0;
	
			} else  if (fabs(value - maximum) < 0.001) {
				multipleBestActions++;
			}
		}
	}

	// if there are multiple best actions, chose a random one 
	if (multipleBestActions) {

		unsigned char found=0;
		unsigned char randAction = rand()%(multipleBestActions+1);
	
		for (_action=0; _action < NUM_ACTIONS; _action++) {
			if (this->env->isValidAction(pos, _action)) {
				if (fabs((value=evaluateAction(pos, _action)) - maximum) < 0.001) {
					if (found < randAction) {
						found++;
					} else {
						bestAction = _action;
						_action = NUM_ACTIONS; // terminates the for-loop
					}
				}
			}
		}
	}
	

	// 2) if there was no best action found, take random action
	if (!haveAction) {	
		bestAction = this->getRandomAction();
		BoardController::blinkLED((1<<LED_EXPLORATION), 50, 200);
	}

	return bestAction;
}


/**
 * this functino performs the value-iteration algorithm for all states
 */
void ValueIteration::valueIterateAllStates() {

	float value, maximum = -100;
	unsigned char bestAction, action, _row, _column;
	unsigned short multipleBestActions = 0;
	bool haveAction = false;
	
	// increment iteration number
	this->iteration++;

	// switch on LEARN LED
	BoardController::powerOnLED(LED_LEARN, true);

	ENV_POS position, d; 
	env->getDiscretization(&d); // discretization


	//BoardController::blinkLED((1<<LED_LEARN), 5, 100);
	
	// for each state s in S: 
	for(_row=0; _row<d.y; _row++) {
		for(_column=0; _column<d.x; _column++) {

			position.x = _column;
			position.y = _row;

			haveAction = false;
			maximum = -100;
			multipleBestActions=0;

			// 1) for each valid successor state
			for(action=0; action<NUM_ACTIONS; action++) {

				if (this->env->isValidAction(&position, action)) {

					value = evaluateAction (&position, action);

					// memorize maximum value
					if(value > maximum) {

						maximum = value;
						haveAction = true;

						// update maximum value in state-space
						//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 2, 300);
						states->setValue(&position, value);
						//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_WALK), 2, 300);
						//BoardController::powerOnLED((1<<LED_EXPLORATION)|(1<<LED_WALK), false);
					}	
				}
			}
		}
	}


	// switch off LEARN LED
	//BoardController::msDelay(100);
	BoardController::powerOnLED(LED_LEARN, false);
}


/**
 * compute the successor-state value for action a in state (_row, _column)
 *   v' = feedback(a) + gamma*v(a)
 */
float ValueIteration::evaluateAction(ENV_POS* pos, unsigned char _action) {

	float neighbourValue, newValue;
	int actionFeedback;

	ENV_POS newPos;
	newPos.x = pos->x;
	newPos.y = pos->y;

	if (_action == UP)   newPos.y--;
	if (_action == DOWN) newPos.y++;
	if (_action == RIGHT) newPos.x++;
	if (_action == LEFT)  newPos.x--;

	// get reward for action a in s
	actionFeedback = states->getFeedback(pos, _action);

	// get value of successor state
	neighbourValue = states->getValue(&newPos);

	// compute new value
	newValue = (gamma * (float)neighbourValue) + (float)actionFeedback;

	return newValue;
}


/**
 * perform a complete exploration of the environment. (all state-transitions)
 */
void ValueIteration::exploreAllFeedbacks(bool wait5Seconds) {

	unsigned char i, j;
	unsigned char steps = 0;

	// switch on EXPLORATION LED
	//BoardController::powerOnLED(LED_EXPLORATION, true);

	// go to state (0,0)
	ENV_POS pos(0, 0);
	env->beamRobotToState(&pos);

	ENV_POS d; // discretization
	env->getDiscretization(&d);

	bool turnFlag = true;

	// delay five seconds
	if (wait5Seconds) BoardController::msDelay(5000);

	//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 4, 400);

	// explore horizontal until (scalX, scalY)
	while (steps < ((d.y*d.x)-1)) {
		
		//BoardController::blinkLED((1<<LED_EXPLORATION), 1, 200);
		// go right
		if (turnFlag) {
	
			moveAndSaveFeedback(RIGHT, false, &pos);
			steps++;

			// go down at the right end
			if (!this->env->isValidAction(&pos, RIGHT)) {
				turnFlag = false;
				if (this->env->isValidAction(&pos, DOWN)) {
					moveAndSaveFeedback(DOWN, false, &pos);
					steps++;
				}
			}

		// go left
		} else {
	
			moveAndSaveFeedback(LEFT, false, &pos);
			steps++;

			// go down on the left end
			if (!this->env->isValidAction(&pos, LEFT)) {
				turnFlag = true;
				if (this->env->isValidAction(&pos, DOWN)) {
					moveAndSaveFeedback(DOWN, false, &pos);
					steps++;
				}
			}
		}
	}

	//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 4, 400);
	// go back horizontally from (scalX, scalY) to (0,0);
	steps=0;
	//turnFlag = !turnFlag;
	while (steps < ((d.y*d.x)-1)) {
		
		//BoardController::blinkLED((1<<LED_EXPLORATION), 1, 200);
		// go right 
		if (turnFlag) {
	
			moveAndSaveFeedback(RIGHT, false, &pos);
			steps++;

			// go up on the right end
			if (!this->env->isValidAction(&pos, RIGHT)) {
				turnFlag = false;
				if (this->env->isValidAction(&pos, UP)){ 
					moveAndSaveFeedback(UP, false, &pos);
					steps++;
				}
			}

		// go left
		} else {
	
			moveAndSaveFeedback(LEFT, false, &pos);
			steps++;
			
			// go up on the left end
			if (!this->env->isValidAction(&pos, LEFT)) {
				turnFlag = true;
				if (this->env->isValidAction(&pos, UP)) {
					moveAndSaveFeedback(UP, false, &pos);
					steps++;
				}
			}
		}		

	}

	//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 4, 400);
	//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 10, 400);
	// explore vertically
	turnFlag = true;
	steps = 0;
	while (steps < ((d.y*d.x)-1)) {
		
		//BoardController::blinkLED((1<<LED_EXPLORATION), 1, 200);

		// go down 
		if (turnFlag) {
	
			moveAndSaveFeedback(DOWN, false, &pos);
			steps++;

			// go left on the lower end
			if (!this->env->isValidAction(&pos, DOWN)) {
				turnFlag = false;
				if (this->env->isValidAction(&pos, RIGHT)) {
					moveAndSaveFeedback(RIGHT, false, &pos);
					steps++;
				}
			}

		// go up
		} else {
	
			moveAndSaveFeedback(UP, false, &pos);
			steps++;

			// go right on the upper end
			if (!this->env->isValidAction(&pos, UP)) {
				turnFlag = true;
				if (this->env->isValidAction(&pos, RIGHT)) {
					moveAndSaveFeedback(RIGHT, false, &pos);
					steps++;
				}
			}
		}
	}

	//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 4, 400);
	//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 10, 400);
	// go vertically backwards to (0,0);
	steps=0;
	while (steps < ((d.y*d.x)-1)) {
		
		//BoardController::blinkLED((1<<LED_EXPLORATION), 1, 200);

		// go down
		if (turnFlag) {
	
			moveAndSaveFeedback(DOWN, false, &pos);
			steps++;
	
			// go left on the lower end
			if (!this->env->isValidAction(&pos, DOWN)) {
				turnFlag = false;
				if (this->env->isValidAction(&pos, LEFT)) {
					moveAndSaveFeedback(LEFT, false, &pos);
					steps++;
				}
			}

		// go left
		} else {
	
			moveAndSaveFeedback(UP, false, &pos);
			steps++;
			
			// go left on the upper end
			if (!this->env->isValidAction(&pos, UP)) {
				turnFlag = true;
				if (this->env->isValidAction(&pos, LEFT)) {
					moveAndSaveFeedback(LEFT, false, &pos);
					steps++;
				}
			}
		}
	}
	//BoardController::blinkLED((1<<LED_LEARN)|(1<<LED_EXPLORATION), 4, 400);


	// turn off FEEDBACK LED
	BoardController::powerOnLED(LED_EXPLORATION, false);
}

/**
 * returns a valid random action 
**/
unsigned char ValueIteration::getRandomAction () {

	bool found = false;
	unsigned char randomAction;
	static unsigned char increment = 0;

	ENV_POS pos;
	env->getCurrentState(&pos);

	do {
		randomAction = (rand()+increment)%NUM_ACTIONS;
		if (this->env-> isValidAction(&pos, randomAction)) { 
			found = true;
		}

		increment++;

	} while (!found);

	return randomAction;
}

/**
 * Go to a random state and update reward table
**/
void ValueIteration::gotoRandomState() {

	ENV_POS d; // discretization
	env->getDiscretization(&d);

	ENV_POS pos;
	env->getCurrentState(&pos);

	unsigned char newX = rand()%d.x;
	unsigned char newY = rand()%d.y;

	// position new state on Y-axis
	while (pos.y != newY) {
		if (pos.y < newY) {
			this->moveAndSaveFeedback(DOWN, 0, &pos);
		}
		else {
			this->moveAndSaveFeedback(UP, 0, &pos);
		}
	}
	// position new state on X-axis
	while (pos.x!= newX) {
		if (pos.x < newX) {
			this->moveAndSaveFeedback(RIGHT, 0, &pos);
		}
		else {
			this->moveAndSaveFeedback(LEFT, 0, &pos);
		}
	}
}

/** returns the next Action with respect to the exploration parameters */
bool ValueIteration::doExploration() {
	
	float explorationProbability = 1;
	float dV;

	if (this->expMode == EXPLORATION_VALUE_BASED) {
		
		// calculate dV
		dV = fabs ((oldValue-newValue) / (fabs(oldValue)+fabs(newValue)));
		explorationProbability = pow(dV, 1.0/this->expParameter);

	} else if (this->expMode == EXPLORATION_EGREEDY) {
		explorationProbability = this->expParameter;
	}

	this->currentExpProbability = explorationProbability;

	// return whether we should explore or exploit
	if ((rand()%100) > (explorationProbability*100)) {
		return false; // exploitation
	} else {
		return true; // exploration
	}
}


/**
 * this method selects the action with respect to the current policy
 */
void ValueIteration::doNextAction() {

	ENV_POS pos, oldPos;

	env->getCurrentState(&pos);
	env->getCurrentState(&oldPos);

	// select next action - with respect on SWTICH_ENABLE_EXPLORATION 
	this->moveAndSaveFeedback( 
		// choose best action
		this->getGreedyAction(&pos),
		// pure greedy or exploitation/exploration
		BoardController::getSwitchState(SWITCH_ENABLE_EXPLORATION),
		// memory for new position
		&pos
	);

	// save old value(currentState) (for VDBE)
	this->oldValue = this->states->getValue(&oldPos); 
	
	// value iterate all states
	this->valueIterateAllStates();

	// save new value(currentState) (for VDBE)
	this->newValue = this->states->getValue(&oldPos);

}


/**
 * perform _action in currentState. _activateRandomMoving defines if exploration/exploitation should be balanced by eGreedy or VDBE. Otherwise, pure greedy exploitation is performed.
 */
char ValueIteration::moveAndSaveFeedback (unsigned char _action, bool _activateRandomMoving, ENV_POS* successorState) {

	ENV_POS oldPos, d;
	env->getCurrentState(&oldPos);
	env->getDiscretization(&d);
	
	//unsigned char oldX = pos.x;
	//unsigned char oldY = pos.y;
	unsigned char randomAction=0;
	unsigned char nextAction = _action;
	
	BoardController::powerOnLED(LED_WALK, true);
	
	// check whether we should explore
	if (_activateRandomMoving && doExploration()) {

		BoardController::powerOnLED(LED_EXPLORATION, true);
		BoardController::powerOnLED(LED_WALK, true);

		if (BoardController::getSwitchState(SWITCH_RANDOM_STATE)==0) {
			nextAction = getRandomAction();
			this->moveAndSaveFeedback(nextAction, 0, successorState);
		} else {
			this->gotoRandomState();
		}

		BoardController::powerOnLED(LED_EXPLORATION, false);
		BoardController::powerOnLED(LED_WALK, false);
		return 0;
	}

	
	//  if _action is invalid, take a (valid) random action
	while (!this->env->isValidAction(&oldPos, nextAction)) {

		BoardController::blinkLED((1<<LED_WALK), 10, 100);
		nextAction = getRandomAction();
	}

	// disable serial interrupt
	//BoardController::disableV24Interrupt();

	// compute delay
	int milliDelay = 0;
	if (nextAction == UP  || nextAction == DOWN) { 
		milliDelay = (int)DELAY_TIME / (int)d.y;
	} else if (nextAction == LEFT || nextAction == RIGHT) {
		milliDelay = (int)DELAY_TIME / (int)d.x;
	}

	// if external memory is used, delay-time can be reduced due to bus latency 
	if (BoardController::getSwitchState(SWITCH_RAM)) {
		milliDelay = (int)((double)milliDelay * 0.66);
	}

	// perform action
	ACTION_RETURN actionReturn;
	this->env->doAction(nextAction, milliDelay, &actionReturn);
	successorState->x = actionReturn.state.x;
	successorState->y = actionReturn.state.y;

	//BoardController::enableV24Interrupt();
	
	// re set seed (with noise from the environment) 
	srand (rand()+1+actionReturn.reward);

	this->distanceCovered += actionReturn.reward;
	states->setFeedback(&oldPos, nextAction, actionReturn.reward);

	// turn of EXPLORATION and WALK LEDs
	BoardController::powerOnLED(LED_EXPLORATION, false);
	BoardController::powerOnLED(LED_WALK, false);

	return actionReturn.reward;
}

unsigned short ValueIteration::get_seed()
{
   unsigned short seed = 0;
   unsigned short *p = (unsigned short*) (RAMEND+1);
   extern unsigned short __heap_start;

   while (p >= &__heap_start + 1)
      seed ^= * (--p);

   return seed;
}

